#!/usr/bin/python

import getopt
import sys

vertices = {}


def key_x(t):
    return t.middle().x


def key_y(t):
    return t.middle().y


def key_z(t):
    return t.middle().z


class Point3D(object):
    def __init__(self, x, y, z):
        self.x = x
        self.y = y
        self.z = z

    def __str__(self):
        return "[%s, %s, %s]" % (self.x, self.y, self.z)


class Primitive(object):
    def __init__(self, primitive_type):
        self.type = primitive_type


class Triangle(Primitive):
    def __init__(self, v1, v2, v3):
        super().__init__('T')
        self.v1 = v1
        self.v2 = v2
        self.v3 = v3

    def middle(self):
        p1 = vertices[self.v1]
        p2 = vertices[self.v2]
        p3 = vertices[self.v3]
        return Point3D((p1.x + p2.x + p3.x) / 3, (p1.y + p2.y + p3.y) / 3, (p1.z + p2.z + p3.z) / 3)

    def __str__(self):
        return "T %s %s %s" % (vertices[self.v1], vertices[self.v2], vertices[self.v3])

    def vertices(self):
        return [self.v1, self.v2, self.v3]


class Box(Primitive):
    def __init__(self, primitives):
        super().__init__('B')
        self.primitives = primitives

    def __add__(self, other):
        return self.primitives + other.primitives


def build_hierarchy(primitives):
    if len(primitives) == 1:
        return primitives[0]
    if len(primitives) <= 2:
        return Box(primitives)
    else:
        # find the min and max values
        min_x = min(vertices[p].x for i in range(len(primitives)) for p in primitives[i].vertices())
        max_x = max(vertices[p].x for i in range(len(primitives)) for p in primitives[i].vertices())
        min_y = min(vertices[p].y for i in range(len(primitives)) for p in primitives[i].vertices())
        max_y = max(vertices[p].y for i in range(len(primitives)) for p in primitives[i].vertices())
        min_z = min(vertices[p].z for i in range(len(primitives)) for p in primitives[i].vertices())
        max_z = max(vertices[p].z for i in range(len(primitives)) for p in primitives[i].vertices())

        # calculate the interval
        x = max_x - min_x
        y = max_y - min_y
        z = max_z - min_z

        if max(x, y, z) == x:
            key = key_x
        elif max(x, y, z) == y:
            key = key_y
        else:
            key = key_z

        # key = (key_x, key_y, key_z)[direction]

        primitives.sort(key=key)
        middle = len(primitives) // 2
        left = primitives[:middle]
        right = primitives[middle:]

        hierarchy_left = build_hierarchy(left)
        hierarchy_right = build_hierarchy(right)

        return Box([hierarchy_left, hierarchy_right])


def write_to_file(primitive, out_file):
    if primitive.type == 'T':
        out_file.write("T %s %s %s\n" % (primitive.v1, primitive.v2, primitive.v3))
    else:
        for i in range(len(primitive.primitives)):
            write_to_file(primitive.primitives[i], out_file)
        out_file.write("B %s\n" % (len(primitive.primitives)))


def main(argv):
    input_file = ''
    output_file = ''
    try:
        opts, args = getopt.getopt(argv, "hi:o:", ["input=", "output="])
    except getopt.GetoptError:
        print('mesh.py -i <input file> -o <output file>')
        sys.exit(2)
    for opt, arg in opts:
        if opt == '-h':
            print('mesh.py -i <input file> -o <output file>')
            sys.exit()
        elif opt in ("-i", "--input"):
            input_file = arg
        elif opt in ("-o", "--output"):
            output_file = arg

    in_file = open(input_file, "r")
    n_vertices = int(in_file.readline())
    for i in range(n_vertices):
        array = in_file.readline().split(' ')
        vertices[i] = Point3D(float(array[0]), float(array[1]), float(array[2]))

    assert n_vertices == len(vertices)

    n_triangles = int(in_file.readline())
    triangles = []
    for i in range(n_triangles):
        array = in_file.readline().split(' ')
        triangles.append(Triangle(int(array[0]), int(array[1]), int(array[2])))

    assert n_triangles == len(triangles)

    bounding_box = build_hierarchy(triangles)

    open(output_file, "w").close()
    out_file = open(output_file, "a")
    out_file.write("%s\n" % n_vertices)
    for i in range(n_vertices):
        out_file.write("%s %s %s\n" % (vertices.get(i).x, vertices.get(i).y, vertices.get(i).z))
    write_to_file(bounding_box, out_file)
    out_file.write("END")


if __name__ == '__main__':
    main(sys.argv[1:])

# python mesh.py -i "C:\Users\alexa\Documents\UCLL\3D computer graphics\bunny.mesh" -o "C:\Users\alexa\Documents\UCLL\3D computer graphics\out.bmesh"
