#include "triangle-primitive.h"

#include "util/misc.h"
#include "math/coordinate-systems.h"
#include "math/quadratic_equation.h"
#include "performance/performance.h"
#include <assert.h>

using namespace raytracer;
using namespace raytracer::primitives;
using namespace math;


namespace
{
    class TriangleImplementation : public raytracer::primitives::_private_::PrimitiveImplementation
    {
    public:
        TriangleImplementation(const math::Point3D& p1, const math::Point3D& p2, const math::Point3D& p3) : m_p1(p1), m_p2(p2), m_p3(p3), m_normal((m_p2 - m_p1).cross(m_p3 - m_p1).normalized() ) {
        }

        std::vector<std::shared_ptr<Hit>> find_all_hits(const Ray& ray) const override
        {
            auto t = ((m_p1 - ray.origin).dot(m_normal)) / (ray.direction.dot(m_normal));
            auto h = ray.origin + ray.direction * t;

            // Allocate vector on stack
            std::vector<std::shared_ptr<Hit>> hits;

            if ((m_p2 - m_p1).cross(h - m_p1).dot(m_normal) >= 0 &&
                (m_p3 - m_p2).cross(h - m_p2).dot(m_normal) >= 0 &&
                (m_p1 - m_p3).cross(h - m_p3).dot(m_normal) >= 0)
            {
                // Create hit object
                auto hit = std::make_shared<Hit>();

                // shared_ptr<T>::get() returns the T* inside the shared pointer
                initialize_hit(hit.get(), ray, t);
                
                // Put hit in list
                hits.push_back(hit);
            }

            return hits;
        }

        math::Box bounding_box() const override
        {
            return Box(interval(std::min(m_p1.x(), std::min(m_p2.x(), m_p3.x())), std::max(m_p1.x(), std::max(m_p2.x(), m_p3.x()))),
                       interval(std::min(m_p1.y(), std::min(m_p2.y(), m_p3.y())), std::max(m_p1.y(), std::max(m_p2.y(), m_p3.y()))),
                       interval(std::min(m_p1.z(), std::min(m_p2.z(), m_p3.z())), std::max(m_p1.z(), std::max(m_p2.z(), m_p3.z()))));
        }

    protected:
        void initialize_hit(Hit* hit, const Ray& ray, double t) const
        {
            hit->t = t;
            hit->position = ray.at(hit->t);
            hit->local_position.xyz = hit->position;
            hit->local_position.uv = Point2D(hit->position.x(), hit->position.y());
            hit->normal = ray.direction.dot(m_normal) < 0 ? m_normal : -m_normal;
        }

        const Point3D m_p1;
        const Point3D m_p2;
        const Point3D m_p3;
        const Vector3D m_normal;
    };
}

Primitive raytracer::primitives::triangle(const math::Point3D& p1, const math::Point3D& p2, const math::Point3D& p3)
{
    return Primitive(std::make_shared<TriangleImplementation>(p1, p2, p3));
}
