#include "raytracers/ray-tracer-v2.h"
#include "easylogging++.h"

using namespace imaging;
using namespace math;
using namespace raytracer;


TraceResult raytracer::raytracers::_private_::RayTracerV2::trace(const Scene& scene, const Ray& ray) const
{
    Hit hit;
    Color result = colors::black();

    // Ask the scene for the first positive hit, i.e. the closest hit in front of the eye
    // If there's a hit, find_first_positive_hit returns true and updates the hit object with information about the hit
    if (scene.root->find_first_positive_hit(ray, &hit))
    {
        // There's been a hit
        // Fill in TraceResult object with information about the trace

        MaterialProperties hit_material = hit.material->at(hit.local_position);

        // This ray tracer always returns white in case of a hit
        result = hit_material.ambient;

        result += process_lights(scene, hit_material, hit, ray);

        // The hit object contains the group id, just copy it (group ids are important for edge detection)
        unsigned group_id = hit.group_id;

        // The t-value indicates where the ray/scene intersection took place.
        // You can use ray.at(t) to find the xyz-coordinates in space.
        double t = hit.t;

        // Group all this data into a TraceResult object.
        return TraceResult(result, group_id, ray, t);
    }
    else
    {
        // The ray missed all objects in the scene
        // Return a TraceResult object representing "no hit found"
        // which is basically the same as returning black
        return TraceResult::no_hit(ray);
    }
}

Color raytracer::raytracers::_private_::RayTracerV2::process_lights(const Scene& scene, const MaterialProperties& matProp, const Hit& hit, const Ray& ray) const
{
    Color result = colors::black();
    for (LightSource l : scene.light_sources) {
        result += process_light_source(scene, matProp, hit, ray, l);
    }
    return result;
}

Color raytracer::raytracers::_private_::RayTracerV2::process_light_source(const Scene& scene, const MaterialProperties& matProp, const Hit& hit, const Ray& ray, LightSource lightSource) const
{
    Color result = colors::black();
    for (LightRay lightray: lightSource->lightrays_to(hit.position)) {
        result += process_light_ray(scene, matProp, hit, ray, lightray);
    }
    return result;
}

Color raytracer::raytracers::_private_::RayTracerV2::process_light_ray(const Scene& scene, const MaterialProperties& matProp, const Hit& hit, const Ray& ray, const LightRay& lightRay) const
{
    Color result = colors::black();
    result += compute_diffuse(matProp, hit, ray, lightRay);
    return result;
}

Color raytracer::raytracers::_private_::RayTracerV2::compute_diffuse(const MaterialProperties& matProp, const Hit& hit, const Ray& ray, const LightRay& lightRay) const
{
    Color result = colors::black();
    Color color;
    double cosAlpha;
    cosAlpha = (lightRay.ray.origin - hit.position).normalized().dot(hit.normal);

    if (cosAlpha > 0) {
        color = lightRay.color * matProp.diffuse;
        result = color * cosAlpha;
    }

    CHECK(result.r >= 0);
    CHECK(result.g >= 0);
    CHECK(result.b >= 0);
    return result;
}

raytracer::RayTracer raytracer::raytracers::v2()
{
    return raytracer::RayTracer(std::make_shared<_private_::RayTracerV2>());
}